/*++
Symbols 32bits Dumper
Copyright 2008 (c) Matthieu Suiche. <msuiche[at]gmail.com>

Sym32Guid is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Sym32Guid is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Sym32Guid.  If not, see <http://www.gnu.org/licenses/>.

Module Name:

    Sym32Guid.c

Abstract:

    - Live Forensics
    - Retrieve PDB, Program Database file from a memory fil.

Environment:

    - User mode

Revision History:

    - Matthieu Suiche

--*/

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

//
// http://www.debuginfo.com/articles/debuginfomatch.html
//
#define CV_SIGNATURE_RSDS 'SDSR'

typedef struct _CV_INFO_PDB70 
{
    DWORD Signature; 
    GUID Guid; // unique identifier 
    DWORD Age; // an always-incrementing value 
    BYTE PdbFileName[1]; // zero terminated string with the name of the PDB file 
} CV_INFO_PDB70, *PCV_INFO_PDB70;

typedef struct _MAPPED_FILE
{
    HANDLE hFile;
    HANDLE hFileMapping;
    PULONG lpFileBase;
} MAPPED_FILE, *PMAPPED_FILE;

PULONG OpenDumpFile(LPCSTR Filename, PMAPPED_FILE MapInfo)
{

    MapInfo->hFile = CreateFileA(Filename, GENERIC_READ, FILE_SHARE_READ, NULL,
                       OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);

    if (MapInfo->hFile == INVALID_HANDLE_VALUE) {
        printf("Error: Couldn't open file with CreateFile()\n");
        return NULL;
    }

    MapInfo->hFileMapping = CreateFileMapping(MapInfo->hFile, NULL, PAGE_READONLY, 0, 0, NULL);
    if (MapInfo->hFileMapping == 0) {
        CloseHandle(MapInfo->hFile);
        printf("Error: Couldn't open file mapping with CreateFileMapping()\n");
        return NULL;
    }

    MapInfo->lpFileBase = (PULONG)MapViewOfFile(MapInfo->hFileMapping, FILE_MAP_READ, 0, 0, 0);
    if (MapInfo->lpFileBase == 0) {
        CloseHandle(MapInfo->hFileMapping);
        CloseHandle(MapInfo->hFile);
        printf("Error: Couldn't map view of file with MapViewOfFile()\n");
        return NULL;
    }

    return MapInfo->lpFileBase;
}

int main(int argc, char **argv)
{
PULONG FileBase;
ULONG HighSize;
ULONG PrintURL, ArgIndex;
PCV_INFO_PDB70 CvInfoPdb70;
MAPPED_FILE MapInfo;
GUID Guid;
ULONG Index;
ULONG RetrievedGUID;

    printf("\n"
            " Sym32GUID - Symbols 32bits GUID dumper.\n"
            " Matthieu Suiche (c) 2008 - http://www.msuiche.net\n\n");
    
    if (argc < 2)
    {
       printf("   Usage: Sym32Guid.exe [option] dumpfile\n");

       printf("Commands:\n"
              "   -u      Give the remote url to download the symbol from Microsoft server.\n\n");

       printf("Sample: \n"
              "   Sym32Guid.exe memory.dump               Lets search into this physical memory dump\n"
              "   Sym32Guid.exe -u memory.dump            Lets search into this physical memory dump AND PRINT URL.\n");

        return 0;
    }

    ArgIndex = 1;
    PrintURL = FALSE;

    if (argv[1][0] == '-')
    {
        PrintURL = TRUE;
        ArgIndex++;
        printf(" !! PLEASE NOTICE THAT USER AGENT MUST BE : !!\n"
               "  User-Agent: Microsoft-Symbol-Server/6.6.0007.5\n\n");
    }

    FileBase = OpenDumpFile(argv[ArgIndex], &MapInfo);
    if (FileBase == NULL) return 0;

    RetrievedGUID = 0;
    printf("Searching for PDB signature....\n\n");

    for (Index = 0; (Index * sizeof(ULONG)) < GetFileSize(MapInfo.hFile, &HighSize); Index++)
    {
        if (FileBase[Index] == CV_SIGNATURE_RSDS)
        {
            CvInfoPdb70 = (PCV_INFO_PDB70)&FileBase[Index];
            Guid = CvInfoPdb70->Guid;
            printf("  Guid: {%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x} ",
                Guid.Data1, Guid.Data2, Guid.Data3, Guid.Data4[0], Guid.Data4[1], Guid.Data4[2],
                Guid.Data4[3], Guid.Data4[4], Guid.Data4[5], Guid.Data4[6], Guid.Data4[7]);
            printf("PDB: %s\n", CvInfoPdb70->PdbFileName);
             if (PrintURL)
            {
                    printf("  GET http://msdl.microsoft.com/download/symbols/%s/%08X%04X%04X%02X%02X%02X%02X%02X%02X%02X%02X%X/%s HTTP/1.0\r\n",
                      CvInfoPdb70->PdbFileName,
                      Guid.Data1, Guid.Data2, Guid.Data3, Guid.Data4[0], Guid.Data4[1], Guid.Data4[2],
                      Guid.Data4[3], Guid.Data4[4], Guid.Data4[5], Guid.Data4[6], Guid.Data4[7],
                      CvInfoPdb70->Age,
                      CvInfoPdb70->PdbFileName);
            }
            RetrievedGUID++;
        }
    }

    printf("[TOTAL:] Sym32GUID retrieved %d GUID signatures.\n", RetrievedGUID);
    //
    // Clean up
    //
    UnmapViewOfFile(FileBase);
    CloseHandle(MapInfo.hFileMapping);
    CloseHandle(MapInfo.hFile);
    return 1;
}